/*
	main.c

	original code: SASANO Takayoshi
	--- public domain, no warranty.
*/

#include "vpd.h"
#include "getmbfid.h"

/* $B%o!<%/%(%j%"(B */
LOCAL	DrvWork	WorkArea;

/* $B%I%i%$%P4XO"%?%9%/5/F0(B */
LOCAL	ERR	all_start(DrvWork *wk)
{
	WERR	err;
	T_CPOR	cpor;
	T_CTSK	ctsk;

	/* $BGSB>@)8fMQ%m%C%/:n@.(B */
	err = CreateLock(&wk->lock);
	if (err < ER_OK) {
		D(("ctl_start: CreateLock %d\n", err));
		goto fin0;
	}

	/* $B%i%s%G%V%]!<%H:n@.(B(ctl) */
	cpor.exinf   = (VP)CH4toW('v', 'p', 'c', 'a' + wk->drvno);
	cpor.poratr  = TA_NULL;
	cpor.maxcmsz = sizeof(DevReq);
	cpor.maxrmsz = sizeof(DevRsp);
	err = vcre_por(&cpor);
	if (err < ER_OK) {
		D(("ctl_start: vcre_por(ctl) %d\n", err));
		goto fin1;
	}
	wk->c_porid = (ID)err;

	/* $B%i%s%G%V%]!<%H:n@.(B(drv) */
	cpor.exinf   = (VP)CH4toW('v', 'p', 'd', 'a' + wk->drvno);
	cpor.poratr  = TA_NULL;
	cpor.maxcmsz = sizeof(DevReq);
	cpor.maxrmsz = sizeof(DevRsp);
	err = vcre_por(&cpor);
	if (err < ER_OK) {
		D(("ctl_start: vcre_por(drv) %d\n", err));
		goto fin2;
	}
	wk->d_porid = (ID)err;

	/* $B%?%9%/:n@.(B(ctl) */
	ctsk.exinf   = wk;
	ctsk.task    = ctl_task;
	ctsk.itskpri = CTLTASK_PRI;
	ctsk.stksz   = CTLTASK_STKSZ;
	ctsk.tskatr  = TA_HLNG | TA_RNG0;
	err = vcre_tsk(&ctsk);
	if (err < ER_OK) {
		D(("ctl_start: vcre_tsk(ctl) %d\n", err));
		goto fin3;
	}
	wk->c_tskid = (ID)err;

	/* $B%?%9%/:n@.(B(drv) */
	ctsk.exinf   = wk;
	ctsk.task    = drv_task;
	ctsk.itskpri = DRVTASK_PRI;
	ctsk.stksz   = DRVTASK_STKSZ;
	ctsk.tskatr  = TA_HLNG | TA_RNG0;
	err = vcre_tsk(&ctsk);
	if (err < ER_OK) {
		D(("ctl_start: vcre_tsk(drv) %d\n", err));
		goto fin4;
	}
	wk->d_tskid = (ID)err;

	/* $B%?%9%/5/F0(B(ctl) */
	err = sta_tsk(wk->c_tskid, D_NORM_PTN | D_ABORT_PTN);
	if (err < ER_OK) {
		D(("ctl_start: sta_tsk(ctl) %d\n", err));
		goto fin5;
	}

	/* $B%?%9%/5/F0(B(drv) */
	err = sta_tsk(wk->d_tskid, D_NORM_PTN | D_ABORT_PTN);
	if (err < ER_OK) {
		D(("ctl_start: sta_tsk(drv) %d\n", err));
		goto fin6;
	}

	err = ER_OK;
	goto fin0;

// fin7:
	ter_tsk(wk->d_tskid);
 fin6:
	ter_tsk(wk->c_tskid);
 fin5:
	del_tsk(wk->d_tskid);
 fin4:
	del_tsk(wk->c_tskid);
 fin3:
	del_por(wk->d_porid);
 fin2:
	del_por(wk->c_porid);
 fin1:
	DeleteLock(&wk->lock);
 fin0:
	return err;
}

/* $B%I%i%$%P=*N;(B */
LOCAL	ERR	all_finish(DrvWork *wk)
{
	if (wk != NULL) {
		del_por(wk->d_porid);	/* $B%?%9%/=*N;MW5a(B(drv) */
		del_por(wk->c_porid);	/* $B%?%9%/=*N;MW5a(B(ctl) */
		Lock(&wk->lock);	/* $B%?%9%/=*N;BT$A(B */
		DeleteLock(&wk->lock);
	}

	return ER_OK;
}

/* $B%I%i%$%P!&%(%s%H%j%]%$%s%H(B */
EXPORT	ERR	main(Bool start, TC *arg)
{
	ERR	err;

	if (!start) {
		err = ER_OK;
		goto fin1;
	}

	/* $B%o!<%/%(%j%"$N=i4|2=(B */
	memset(&WorkArea, 0, sizeof(WorkArea));
	WorkArea.drvno = tc_islower(*arg) ? (*arg - TK_a) : 0;
	WorkArea.mbfid = get_mbfid();

	/* $B4IM}%?%9%/$N5/F0(B */
	err = all_start(&WorkArea);
	if (err < ER_OK) goto fin0;

	err = ER_OK;
	goto fin0;

 fin1:
	all_finish(&WorkArea);
 fin0:
	return err;
}
