/*
	vpdtool.c

	original code: SASANO Takayoshi
	--- public domain, no warranty.
*/

#include <btron/btron.h>
#include <device/disk.h>
#include <tcode.h>
#include <errcode.h>
#include <bstdio.h>
#include <bstdlib.h>
#include <bstring.h>
#include <tstring.h>
#include "vpdpart.h"

/* buffer */
LOCAL	UB		Buf[128];
LOCAL	VPD_PARTINFO	VPD_PartInfo[MAX_PARTINFO];
LOCAL	VPD_PARTINFO	VPD_Temp[MAX_PARTINFO];

/* choice entry */
LOCAL	W	choice_entry(W entry)
{
	W	err, e;

	printf("entry (1-%d): ", entry);
	gets(Buf);
	e = atoi(Buf);
	if (e < 1 || e > entry) {
		err = ER_PAR;
		goto fin0;
	}
	err = e - 1;

 fin0:
	return err;
}

/* choice proceed or not */
LOCAL	W	choice_proceed(void)
{
	printf("proceed (y/[n]) : ");
	gets(Buf);

	return (*Buf == 'y') ? ER_OK : ER_PAR;
}

/* modify patition table without error check */
LOCAL	W	edit_modify(W entry, DiskInfo *di)
{
	W	err, e;

	err = choice_entry(entry);
	if (err < ER_OK) goto fin0;
	e = err;

	/* modify entry */
	printf("systemid  (0x%08x): ", VPD_PartInfo[e].systemid);
	gets(Buf);
	VPD_PartInfo[e].systemid = atoi(Buf);

	printf("startblock(%10d): ", VPD_PartInfo[e].startblock);
	gets(Buf);
	VPD_PartInfo[e].startblock = atoi(Buf);

	printf("endblock  (%10d): ", VPD_PartInfo[e].endblock);
	gets(Buf);
	VPD_PartInfo[e].endblock = atoi(Buf);

	err = ER_OK;
 fin0:
	return err;
}

/* create new table with error check */
LOCAL	W	edit_create(W entry, DiskInfo *di)
{
	W	err, e, start, size;

	err = choice_entry(entry);
	if (err < ER_OK) goto fin0;
	e = err;

	/* make copy */
	memcpy(VPD_Temp, VPD_PartInfo, sizeof(VPD_PartInfo));

	if (VPD_Temp[e].systemid > 0) {
		printf("already used\n");
		err = ER_BUSY;
		goto fin0;
	}

	VPD_Temp[e].systemid = DSID_BTRON;

	printf("startblock: ");
	gets(Buf);
	start = atoi(Buf);
	VPD_Temp[e].startblock = start;

	printf("partition size (MB): ");
	gets(Buf);
	size = atoi(Buf) * (1024 * 1024 / di->blocksize);
	VPD_Temp[e].endblock = start + size - 1;

	/* check */
	err = check_partinfo(VPD_Temp, MAX_PARTINFO, 1, di->blockcont - 1);
	if (err < ER_OK) {
		printf("bad value\n");
		goto fin0;
	}

	/* update */
	memcpy(VPD_PartInfo, VPD_Temp, sizeof(VPD_Temp));

	err = ER_OK;
 fin0:
	return err;
}

/* delete partition table */
LOCAL	W	edit_delete(W entry)
{
	W	err, e;

	err = choice_entry(entry);
	if (err < ER_OK) goto fin0;
	e = err;

	err = choice_proceed();
	if (err < ER_OK) goto fin0;

	VPD_PartInfo[e].systemid = VPD_PartInfo[e].startblock =
		VPD_PartInfo[e].endblock = VPD_PartInfo[e].reserved = 0;

	err = ER_OK;
 fin0:
	return err;
}

/* edit: main loop */
LOCAL	W	edit_main(DiskInfo *di, W entry)
{
	W	err, i;

	while (1) {
		printf("*** blocksize: %d  blockcont: %d\n",
		       di->blocksize, di->blockcont);

		for (i = 0; i < entry; i++) {
			printf("[%2d] (0x%08x) %10d -> %10d\n", i + 1,
			       VPD_PartInfo[i].systemid,
			       VPD_PartInfo[i].startblock,
			       VPD_PartInfo[i].endblock);
		}

		printf("cmd: ");
		gets(Buf);

		switch (*Buf) {
		case	'e':
			err = ER_OK;
			goto fin0;
		case	'q':
			err = ER_PAR;
			goto fin0;
		case	'c':
			edit_create(entry, di);
			break;
		case	'm':
			edit_modify(entry, di);
			break;
		case	'd':
			edit_delete(entry);
			break;
		default:
			printf("e - save and exit\n");
			printf("q - quit without save\n");
			printf("c - create new table entry\n");
			printf("m - modify table entry\n");
			printf("d - delete table entry\n");
			break;
		}

		printf("\n");
	}
 fin0:
	return err;
}

/* edit */
LOCAL	W	cmd_edit(W ac, TC *av[])
{
	W	err, dd, rsz, entry;
	UB	*buf;
	DiskInfo	di;
	VPD_HEADER	*h;

	/* clear buffer */
	memset(VPD_PartInfo, 0, sizeof(VPD_PartInfo));
	memset(VPD_Temp, 0, sizeof(VPD_Temp));

	/* check argument */
	if (ac < 3) {
		printf("usage:	%S e [target device]\n", av[0]);
		err = ER_PAR;
		goto fin0;
	}

	/* open target device */
	err = opn_dev(av[2], D_UPDATE | D_EXCL, NULL);
	if (err < ER_OK) {
		printf("cmd_edit: opn_dev %d\n", err);
		goto fin0;
	}
	dd = err;

	/* check DiskInfo */
	err = rea_dev(dd, DN_DISKINFO, (UB *)&di, sizeof(di), &rsz, NULL);
	if (err < ER_OK) {
		printf("cmd_edit: rea_dev %d\n", err);
		goto fin1;
	}

	/* allocate buffer */
	buf = malloc(di.blocksize);
	if (buf == NULL) {
		printf("cmd_edit: malloc NULL\n");
		err = ER_NOMEM;
		goto fin2;
	}

	/* load 1st sector */
	err = rea_dev(dd, 0, buf, 1, &rsz, NULL);
	if (err < ER_OK) {
		printf("cmd_edit: rea_dev %d\n", err);
		goto fin2;
	}

	/* check header */
	h = check_header(buf, MAX_PARTINFO, di.blocksize);
	if (h == NULL) {
		/* invalid header: create new header */
		printf("header corrupted, create new table\n");
		err = choice_proceed();
		if (err < ER_OK) goto fin2;
		err = create_header(buf, MAX_PARTINFO, di.blocksize);
		if (err < ER_OK) goto fin2;
	}

	/* import */
	err = import_partinfo(buf, VPD_PartInfo, MAX_PARTINFO, di.blocksize);
	if (err < ER_OK) {
		printf("cmd_edit: import_partinfo %d\n", err);
		goto fin2;
	}
	entry = err;

	/* check table entries (no error check) */
	check_partinfo(VPD_PartInfo, entry, 0, di.blockcont - 1);

	/* edit table entries */
	err = edit_main(&di, entry);
	if (err < ER_OK) {
		goto fin2;
	}

	/* export */
	err = export_partinfo(buf, VPD_PartInfo, entry, di.blocksize);
	if (err < ER_OK) {
		printf("cmd_edit: export_partinfo %d\n", err);
		goto fin2;
	}

	/* save 1st sector */
	err = wri_dev(dd, 0, buf, 1, &rsz, NULL);
	if (err < ER_OK) {
		printf("cmd_edit: wri_dev %d\n", err);
		goto fin2;
	}

	err = ER_OK;
 fin2:
	free(buf);
 fin1:
	cls_dev(dd, 0, NULL);
 fin0:
	return err;
}

/* attach */
LOCAL	W	cmd_attach(W ac, TC *av[])
{
	W	err, dd, rsz;

	/* check argument */
	if (ac < 4) {
		printf("usage:	%S a [vpdctl device] [target device]\n",
		       av[0]);
		err = ER_PAR;
		goto fin0;
	}

	/* open vpdctl device */
	err = opn_dev(av[2], D_UPDATE | D_EXCL, NULL);
	if (err < ER_OK) {
		printf("cmd_attach: opn_dev %d\n", err);
		goto fin0;
	}
	dd = err;

	/* attach */
	err = wri_dev(dd, DN_VPDATTACH, (UB *)av[3],
		      (tc_strlen(av[3]) + 1) * sizeof(TC), &rsz, NULL);
	if (err < ER_OK) {
		printf("cmd_attach: wri_dev %d\n", err);
		goto fin1;
	}

	err = ER_OK;
 fin1:
	cls_dev(dd, 0, NULL);
 fin0:
	return err;
}

/* detach */
LOCAL	W	cmd_detach(W ac, TC *av[])
{
	W	err, dd, rsz;

	/* check argument */
	if (ac < 3) {
		printf("usage:	%S d [vpdctl device]\n", av[0]);
		err = ER_PAR;
		goto fin0;
	}

	/* open vpdctl device */
	err = opn_dev(av[2], D_UPDATE | D_EXCL, NULL);
	if (err < ER_OK) {
		printf("cmd_detach: opn_dev %d\n", err);
		goto fin0;
	}
	dd = err;

	/* detach */
	err = wri_dev(dd, DN_VPDDETACH, NULL, 0, &rsz, NULL);
	if (err < ER_OK) {
		printf("cmd_detach: wri_dev %d\n", err);
		goto fin1;
	}

	err = ER_OK;
 fin1:
	cls_dev(dd, 0, NULL);
 fin0:
	return err;
}

/* status */
LOCAL	W	cmd_status(W ac, TC *av[])
{
	W	err, dd, rsz;

	/* check argument */
	if (ac < 3) {
		printf("usage:	%S s [vpdctl device]\n", av[0]);
		err = ER_PAR;
		goto fin0;
	}

	/* open vpdctl device */
	err = opn_dev(av[2], D_UPDATE | D_EXCL, NULL);
	if (err < ER_OK) {
		printf("cmd_detach: opn_dev %d\n", err);
		goto fin0;
	}
	dd = err;

	memset(Buf, 0, sizeof(Buf));

	/* check status */
	err = rea_dev(dd, DN_VPDATTACH, Buf, sizeof(Buf), &rsz, NULL);
	if (err < ER_OK) {
		printf("not attached\n");
	} else {
		printf("attached: %S\n", Buf);
	}

	err = ER_OK;
 fin1:
	cls_dev(dd, 0, NULL);
 fin0:
	return err;
}

/* main */
EXPORT	ERR	main(W ac, TC *av[])
{
	W	err;

	if (ac < 2) {
		printf("usage:	%S [cmd] [args...]\n", av[0]);
		err = ER_PAR;
		goto fin0;
	}

	switch (*av[1]) {
	case	TK_a:
		err = cmd_attach(ac, av);
		break;
	case	TK_d:
		err = cmd_detach(ac, av);
		break;
	case	TK_e:
		err = cmd_edit(ac, av);
		break;
	case	TK_s:
		err = cmd_status(ac, av);
		break;
	default:
		printf("command:\n");
		printf("a - attach\n");
		printf("d - detach\n");
		printf("e - edit\n");
		printf("s - status\n");
		err = ER_PAR;
		break;
	}

 fin0:
	return err;
}
